/***************************************************************************
 *   Copyright (C) 2009 by Guy Rutenberg <guyrutenberg@gmail.com>	   *
 *                                                                         *
 *   This is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "basic_ai_engine.h"
#define INVALID_MOVE (-1)

BasicAiEngine::BasicAiEngine(int search_depth)
{
	scoring_house[0] = scoring_house[1] = 0;

	for (int i = 0 ; i < 12; i++)
		house_[i] = 4;
	
	this->search_depth = search_depth;
	last_move = 0;
	next_player = 0;
}

int BasicAiEngine::scoringHouse(int i)
{
	if (i==0 || i==1)
		return scoring_house[i];
	return -1;
}

int BasicAiEngine::house(int i)
{
	if (0<=i && i<12)
		return house_[i];
	return -1;
}

int BasicAiEngine::lastMove()
{
	return last_move;
}

bool BasicAiEngine::makeMove(int i)
{
	if (!isValidMove(i))
		return false;
	makeUnvalidatedMove(i);
	return true;
}

void BasicAiEngine::makeUnvalidatedMove(int i)
{
	int other_player = (next_player+1) % 2;
	int j = i;
	while (house_[i]) {
		j = (j + 1) % 12;
		// skip over the original house when seeding
		if (i == j)
			j = (j + 1) % 12;
		house_[i]--;
		(house_[j])++;
	}

	//now j holds the position of the last seed sowed.
	bool capture[6] = {0};
	
	//The second condition ensures we capture only on opponent's side
	if (((house_[j] == 2 || house_[j] == 3) ) && \
	  ((other_player*6 <= j) && (j < (other_player*6 + 6))))
		capture[j%6] = true;

	while (capture[j%6] && j%6) {
		j--;
		if (house_[j] == 2 || house_[j] == 3)
			capture[j%6] = true;
	}

	bool all_seeds_captured = true;
	for (j = other_player*6; j < (other_player*6 + 6); j++) {
		if (!capture[j%6] && house_[j])
			all_seeds_captured = false;
	}

	// capture the seeds
	if (!all_seeds_captured) {
		for (j = other_player*6; j < (other_player*6 + 6); j++) {
			if (capture[j%6]) {
				scoring_house[next_player] += house_[j];
				house_[j] = 0;
			}
		}
	}

	bool player_side_empty = true;
	for (j = next_player*6; j<(next_player*6+6); j++) {
		if (house_[j]) {
			player_side_empty = false;
			break;
		}
	}

	// check if opponent can give player some seeds
	if (player_side_empty) {
		bool opponent_can_give_seeds = false;
		for (j = other_player*6; j < (other_player*6 + 6); j++) {
			if (house_[j] >= (other_player+1)*6 - j) {
				opponent_can_give_seeds = true;
				break;
			}
		}
		//if not all remaining seeds go to the opponent
		if (opponent_can_give_seeds) {
			for (j = other_player*6; j < (other_player*6 + 6); j++) {
				scoring_house[other_player] += house_[j];
				house_[j] = 0;
			}
		}
	}

	next_player = other_player;
	last_move = i;
}

int BasicAiEngine::nextPlayer()
{
	return next_player;
}


int BasicAiEngine::makeAiMove()
{
	int heuristic_value;
	int move;
	move = alphaBetaPrune(search_depth, -200, 200, heuristic_value);
	makeUnvalidatedMove(move);

	return heuristic_value;
}

bool BasicAiEngine::isValidMove(int i)
{
	if ((next_player*6 > i) || (i >= (next_player*6 + 6))) {
		// i not in range for current player.
		return false;
	}
	
	if (!house_[i]) {
		// no seeds to sow
		return false;
	}

	// Check whether the others player side is empty, and if so make sure
	// the move gives him some seeds to play with
	int other_player = (next_player+1) % 2;
	int sum = 0;
	for(int j = other_player*6; j < (other_player*6 + 6); j++) {
		sum += house_[j];
	}
	if (sum == 0 && (house_[i] < (next_player+1)*6 - i)) {
		return false;
	}

	return true;
}

bool BasicAiEngine::isGameOver()
{
	for (int i = 0; i < 12; i++) {
		if (house_[i])
			return false;
	}
	return true;
}

//
/**
 * Make an Alpha-Beta prunning to find the next best move.
 * \param depth How deep to search the game-tree.
 * \param heuristic_value Returns the heuristic value associated with the current position.
 * \returns The best move according to the Alpha-Beta algorithm.
 */
int BasicAiEngine::alphaBetaPrune(int depth, int alpha, int beta, int &heuristic_value)
{
	if (depth == 0 || isGameOver()) {
		heuristic_value = heuristicEvaluation();
		return INVALID_MOVE;
	}

	int child_heuristics;
	int best_move;

	for (int move = 0; move < 12; move++) {
		// skip over invalid moves
		if (!isValidMove(move))
			continue;

		BasicAiEngine temp = *this;
		temp.makeUnvalidatedMove(move);
		temp.alphaBetaPrune(depth - 1, alpha, beta, child_heuristics);

		if (next_player) { //it's second player turn - Min
			if (beta > child_heuristics) {
				beta = child_heuristics;
				best_move = temp.last_move;
			}
			
			if (alpha >= child_heuristics)
				break;
		} else { // it's the first player turn - Max
			if (alpha < child_heuristics) {
				alpha = child_heuristics;
				best_move = temp.last_move;
			}
			
			if (beta <= child_heuristics)
				break;
		}

	}

	heuristic_value = next_player ? beta : alpha;
	return best_move;
}

/**
 * \returns A heuristic evaluation of the current game state
 */
int BasicAiEngine::heuristicEvaluation()
{
	int players_score[2];
	for (int i = 0; i < 2; i++) {
		players_score[i] = scoring_house[i];
		if (players_score[i] >=24)
			players_score[i] += 100;
	}
	return players_score[0] - players_score[1];
}
